################# A Model of QE, Reserve Demand, and the Money Multiplier #################
################# Authors: Ellen Ryan and Karl Whelan #################
  
##############File overview and pre-requisites 

### No pre-requisites
## File contains two functions: 1st is basic model, second iterates X number of times
## This file must be run before files 2 and 3.2 which produce model output

#################Input variables for first model:

# interation - keep at one

# cb - vector of central bank purchases/ changes in monetary base. Length of this vector will determine how many periods model runs for

# beta_h, beta_l, sh_low_beta - model can include 2 beta values but paper just uses beta_h and sets share beta_l to zero (or setting them the same)

# num_banks - number of banks in the system

# sh_seller - share of banking system which EVER receive reserves from the central bank

# sh_seller_pp - share of banking system receive reserves from the central bank IN EACH PERIOD - must be <= sh_seller

# res_start, dep_start - starting aggregate reserve and deposit values. Starting aggregate loans is difference between 2

# non_bank_sh - share of sellers to CB who are non-banks

# curr_res - vector giving currency to RESERVE ratio for each period IN THIS MODEL MUST BE VECTOR SAME LENGTH AS CB

# repay - monthly repayment rate on assets

# print_detail - set to TRUE if you want R to print out what is happening in underlying model - slows down process but helpful for troubleshooting



#############Input variables for second model:

# num_it - how many times do you want the model to be repeated?

# i_[variable name] - see equivalent variable above 


##################################### Core model sets out hot potato dynamic #################################

mm_model_core_new <-function(iteration = iteration, cb = cb, beta_h = beta_h, beta_l = beta_h, num_banks = num_banks, 
                                 sh_low_beta = sh_low_beta, sh_seller = sh_seller, seller_pp = seller_pp,
                                 res_start = res_start, dep_start = dep_start, curr_res = curr_res, 
                                 non_bank_sh = non_bank_sh, repay = repay,
                                 print_detail = print_detail){
  
  
  ##################################### Starting points #############################################
  
  #Reserves balance sheet entries
  res_bs <- rep(res_start/num_banks,num_banks)
  #Deposit balance sheet entries
  dep_bs <- rep((dep_start/num_banks),num_banks)
  #Loan balance sheet entries
  loan_bs <- rep((dep_start-res_start)/num_banks,num_banks)
  
  currency<- curr_res[1]*res_start
  
  
  #there is a stronger reaction by beta_h banks
  #share of high and low betas determined by model input - randomly assigned
  #at start and stay the same throughout
  num_low_beta<-round(num_banks*sh_low_beta,0)
  beta_vec<-c(rep(beta_l,num_low_beta),rep(beta_h,num_banks-num_low_beta))
  beta_vec<-beta_vec[sample(1:num_banks, replace = F)]
  
  # Set up output dataframes - model will append new row to these each period. Key parameters are
  # included but should also be included in final object name to be sure that output with different parameter
  # values is easily identifiable
  start_panel <- data.frame(beta = beta_vec,
                            sh_low_beta = rep(sh_low_beta,num_banks),
   #                         seller = c(rep(1,num_banks*sh_seller),rep(0,num_banks*(1-sh_seller))),
                            iteration = rep(iteration, num_banks),
                            period = rep(0,num_banks),
                            bank = seq(1,num_banks),
                            reserves = res_bs,
                            deposits = dep_bs,
                            loans = loan_bs,
                            outflow = rep(0,num_banks))
  
  start_stats<- data.frame(beta_h = beta_h,
                           beta_l = beta_l,
                           sh_low_beta = sh_low_beta,
                           sh_seller = sh_seller,
                           num_banks = num_banks,
                           iteration = iteration,
                           period = 0,
                           cb = 0,
                           bank_push = 0,
                           tot_net_flow = 0,
                           gross_flow = 0,
                           total_deposits = sum(dep_bs),
                           total_reserves = sum(res_bs),
                           total_loans = sum(loan_bs),
                           currency = currency,
                           share_base_ch_curr = 0,
                           model_money_mult = sum(c(dep_bs, currency)) / sum(c(res_bs, currency)),
                           model_nc_mult = sum(c(dep_bs)) / sum(c(res_bs)),
                           loan_repay = 0)
  
  
  if(iteration == 1){
    bank_panel <- start_panel
    key_stats <- start_stats
  }
  
  if(iteration > 1){
    bank_panel <- rbind(bank_panel, start_panel)
    key_stats <- rbind(key_stats, start_stats)
  }
  
  
  ###################TIME LOOP START
  
  
  period <- 1
  
  repeat{
    if (period <= length(cb)) {
      if(print_detail == TRUE){print(c("Period ", period))}
      
      
      ################## What share of base expansion is currency v reserves? ###############
      
      
      #When central bank changes monetary base (cb), what change in currency will return
      #currency to reserve ratio to what we want? 
      
      currency_eq <- (curr_res[period]/(curr_res[period]+1))*(sum(res_bs)+currency+cb[period])
      
      currency_ch<- currency_eq-currency
      
      #Currency adjusts to equilibrium value
      currency<-currency_eq
      
      #This comes out of this period's central bank activity
      cb[period]<- cb[period] - currency_ch
      
      if(cb[period] == 0){share_base_ch_curr <- 0 }else{
      share_base_ch_curr <- currency_ch/(cb[period] + currency_ch)}
      
      
      if(print_detail == TRUE){print("Currency change")}
      if(print_detail == TRUE){print(currency_ch)}     
      
      if(print_detail == TRUE){print("Share of base expansion which is currency")}
      if(print_detail == TRUE){print(currency_ch/(cb[period] + currency_ch))}    
      
      if(print_detail == TRUE){print("New reserves from central bank purchases")}
      if(print_detail == TRUE){print(cb[period])}     
      
      ##################################### Central bank purchases #######################################
      
      #Central bank buys bonds from a share of banks in each given period (seller_pp)
      #if sh_seller < 1 only a subset of banks ever act as sellers
      selling_sample<-round(num_banks*sh_seller,0)
      number_sellers<-round(num_banks*seller_pp,0)
      cb_per_bank<-cb[period]/number_sellers
      x<-round(sample(1:selling_sample,number_sellers, replace = F),0)

      res_bs[x]<- res_bs[x]+cb_per_bank
      dep_bs[x]<- dep_bs[x]+cb_per_bank*non_bank_sh
      loan_bs[x]<- loan_bs[x]-cb_per_bank*(1-non_bank_sh)
      if(print_detail == TRUE){print("Reserve balances after CB purch") }
      if(print_detail == TRUE){print(res_bs)}
      
      
      
      ######################################### Loan repayments #########################################
      
      ### A share of outstanding loans are repayed via a reduction in deposits,
      ### movement of reserves between banks and a reduction in loans
      
       tot_repay <- repay*(sum(loan_bs))
      
      if(print_detail == TRUE){print(c("Total loan repayments value"))}
      if(print_detail == TRUE){print(tot_repay)}            
      
      ### Deposit reductions are distributed in line with each banks'
      ## share of total deposits in each bank, same with loan reductions
      
      repay_dep <- tot_repay * dep_bs/(sum(dep_bs))
      
      if(print_detail == TRUE){print(c("Bank -level deposit reductions"))}
      if(print_detail == TRUE){print(repay_dep)}
      
      check <- tot_repay - sum(repay_dep)
      
       if(print_detail == TRUE){print(c("Check this equals total - difference equals zero?"))}
       if(print_detail == TRUE){print(check)}
      
       dep_bs <- dep_bs - repay_dep
      
       res_bs <- res_bs - repay_dep
      
       repay_loan <- tot_repay * loan_bs/(sum(loan_bs))
      
      
        if(print_detail == TRUE){print(c("Bank -level loan reductions"))}
        if(print_detail == TRUE){print(repay_loan)}
      
        check <- tot_repay - sum(repay_loan)
      
        if(print_detail == TRUE){print(c("Check this equals total  - difference equals zero?"))}
        if(print_detail == TRUE){print(check)}
      
      loan_bs <- loan_bs - repay_loan
      
      res_bs <- res_bs + repay_loan
      
      net_effect <- repay_loan - repay_dep
      
      if(print_detail == TRUE){print(c("Bank-level net effect on reserve balances"))}
      if(print_detail == TRUE){print(net_effect)}  
      
      ############################  Banks' reserve management #######################################
      
      
      #All banks decide to push excess reserves
      #off their balance sheets -
      
      #Banks can't give themselves a negative reserve balance
      outflow1 <- res_bs - (1-beta_vec)*dep_bs
      outflow<-ifelse(outflow1>res_bs,res_bs,outflow1)
      outflow<-ifelse(outflow1<0,0,outflow)
      

      if(print_detail == TRUE){print("Outflows")}
      if(print_detail == TRUE){print(outflow)}
      res_bs <- res_bs - outflow
      if(print_detail == TRUE){print("Reserve balances after outflows")}
      if(print_detail == TRUE){print(res_bs)}
      
      
      
      #These land on the balance sheets of random banks
      #On the asset side as reserves and on the deposit side
      #as deposits, they are transferred via the creation of loans
      inflow <- sample(outflow)
      if(print_detail == TRUE){print("Inflows")}
      if(print_detail == TRUE){print(inflow)}
      res_bs <- res_bs + inflow
      if(print_detail == TRUE){print("Reserve balances after inflows") }
      if(print_detail == TRUE){print(res_bs)}
      dep_bs <- dep_bs + inflow
      loan_bs <- loan_bs + outflow
      if(print_detail == TRUE){print("Deposit balances after inflows")}
      if(print_detail == TRUE){print(dep_bs)}
      if(print_detail == TRUE){print("Loan balances after outflows")}
      if(print_detail == TRUE){print(loan_bs)}
      
      
      
      #For each bank what is net result of hot potato effect?
      net_flow <- inflow-outflow
      if(print_detail == TRUE){print("Net flow per bank excl. reserves from CB sales")}
      if(print_detail == TRUE){print(net_flow)}
      
      
      ##################################### End of period stats ##################################
      
      #What is the change in reserves for each bank over the period?
      
      res_ch_abs<- abs(res_bs - tail(bank_panel$reserves,n=num_banks))
      
      if(print_detail == TRUE){print("Reserves change incl. CB sales - absolute values")}
      if(print_detail == TRUE){print(res_ch_abs)}
     
      
       #Calculate key statistics
      new_stats<-c(beta_h = beta_h,
                   beta_l = beta_l,
                   sh_low_beta = sh_low_beta,
                   sh_seller = sh_seller,
                   num_banks = num_banks,
                   iteration = iteration,
                   period = period,
                   cb = cb[period],
                   bank_push = sum(outflow),
                   tot_net_flow = cb[period],
                   gross_flow = sum(res_ch_abs),
                   total_deposits = sum(dep_bs),
                   total_reserves = sum(res_bs),
                   total_loans = sum(loan_bs),
                   currency = currency,
                   share_base_ch_curr = share_base_ch_curr,
                   model_money_mult = sum(c(dep_bs, currency)) / sum(c(res_bs, currency)),
                   model_nc_mult     = sum(c(dep_bs)) / sum(c(res_bs)),
                   loan_repay = tot_repay)
      key_stats<-rbind(key_stats,new_stats)
      
      
      
      
      #Add period end reserve balances to create panel (with some key stats for 
      #model output identification)
      y<-data.frame(beta = beta_vec,
                    sh_low_beta = sh_low_beta,
            #        seller = c(rep(1,num_banks*sh_seller),rep(0,num_banks*(1-sh_seller))),
                    iteration = rep(iteration,num_banks),
                    period =rep(period,num_banks),
                    bank = seq(1,num_banks),
                    reserves = res_bs,
                    deposits = dep_bs,
                    loans = loan_bs,
                    outflow = outflow)
      bank_panel<-rbind(bank_panel,y)
      bank_panel$bank<-as.factor(bank_panel$bank)
      
      
      
      
      #####################Start next period or end
      
      period <- period + 1
    } else {break}
    
    #Save panel and key statistics outside of model
    bank_panel<<-bank_panel
    key_stats<<-key_stats
    
  }}


##################################################### Demo run #################################################


#NOTE: curr_res AND CB MUST BE SAME LENGTH

mm_model_core_new(iteration = 1, cb = rep(5,3), beta_h = 1.01, beta_l = 0.99, num_banks = 10, 
                        sh_low_beta = 0, sh_seller = 0.8, seller_pp = 0.6,
                        res_start = 1, dep_start = 100, curr_res = c(rep(10,3)), repay = 0.02,
                        non_bank_sh = 0.75,
                        print_detail = TRUE)


#Install package for charts if necessary
#install.packages("ggplot2")
library(ggplot2)

ggplot() + geom_line(data = key_stats,  aes(x = period, y = model_money_mult)) 
ggplot() + geom_line(data = key_stats,  aes(x = period, y = currency)) 
ggplot() + geom_line(data = key_stats,  aes(x = period, y = currency/total_deposits)) 
ggplot() + geom_line(data = key_stats,  aes(x = period, y = currency/total_reserves)) 
ggplot() + geom_line(data = key_stats,  aes(x = period, y = share_base_ch_curr)) 






##################### Iterations model takes core model and runs it X number of times##########################


mm_model_core_new_iterations <-function(num_it, i_cb,i_beta_h, i_beta_l = i_beta_h, i_num_banks, 
                                      i_sh_low_beta = 0, i_sh_seller,
                                      i_seller_pp,
                                      i_res_start, i_dep_start, 
                                      i_curr_res,
                                      i_non_bank_sh, i_repay,
                                      i_print_detail = FALSE){
  
  ################### Run first iteration to set up dataframes
  
  mm_model_core_new(iteration = 1, cb = i_cb,beta_h = i_beta_h, beta_l = i_beta_l,num_banks = i_num_banks, 
                sh_low_beta = i_sh_low_beta, sh_seller = i_sh_seller,
                seller_pp = i_seller_pp,
                res_start = i_res_start, dep_start = i_dep_start, 
                curr_res = i_curr_res,
                non_bank_sh = i_non_bank_sh, repay = i_repay,
                print_detail = i_print_detail)
  
  ################# Use apply function to do all other iterations
  
  iteration <- seq(2,num_it)
  
  
  lapply(iteration,mm_model_core_new, cb = i_cb,beta_h = i_beta_h, beta_l = i_beta_l,num_banks = i_num_banks, 
         sh_low_beta = i_sh_low_beta, sh_seller = i_sh_seller,
         seller_pp = i_seller_pp,
         res_start = i_res_start, dep_start = i_dep_start, 
         curr_res = i_curr_res,
         non_bank_sh = i_non_bank_sh, repay = i_repay,
         print_detail = i_print_detail)

}





